from typing import Dict

from torch import clamp, nn, Tensor

from pi_vae_pytorch.decoders import GINFlowDecoder
from pi_vae_pytorch.encoders import MLPEncoder
from pi_vae_pytorch.layers import ZPriorContinuous, ZPriorDiscrete
from pi_vae_pytorch.utils import compute_posterior, generate_latent_z


class PiVAE(nn.Module):
    """
    The pi-VAE model which constructs a latent variable model of neural activity while 
    simultaneously modeling the relation between the latent and task variables.

    Parameters
    ----------
    x_dim (int) - observed x dimension
    u_dim (int) - 
    z_dim (int) - latent z dimension
    discrete_labels (int) - label u format discrete/continuous Default: True
    encoder_n_hidden_layers (int) - number of the MLPEncoder's MLP hidden layers. Default: 2
    encoder_hidden_layer_dim (int) - dimension of the MLPEncoder's MLP hidden layers. Default: 120
    encoder_hidden_layer_activation (nn.Module) - activation function applied to the MLPEncoder's MLP hidden layers. Default: nn.Tanh
    decoder_n_gin_blocks (int) - number of GINBlocks in the GINFlowDecoder. Default: 2
    decoder_gin_block_depth (int) - depth of each GINBlock in the GINFlowDecoder. Default: 2
    decoder_affine_input_layer_slice_dim (int) - index at which to split an n-dimensional sample x input to each AffineCouplingLayer. Default: None
    decoder_affine_n_hidden_layers (int) - number of each AffineCouplingLayer's MLP hidden layers. Default: 2
    decoder_affine_hidden_layer_dim (int) - dimension of each AffineCouplingLayer's MLP hidden layers. Default: None
    decoder_affine_hidden_layer_activation (nn.Module) - activation function applied to each AffineCouplingLayer's MLP hidden layers. Default: nn.ReLU
    decoder_nflow_n_hidden_layers (int) - number of the NFlowLayer's MLP hidden layers. Default: 2
    decoder_nflow_hidden_layer_dim (int) - dimension of the NFlowLayer's MLP hidden layers. Default: None
    decoder_nflow_hidden_layer_activation (nn.Module) - activation function applied to the NFlowLayer's MLP hidden layers. Default: nn.ReLU
    decoder_observation_model (str) - GINFlowDecoder's observation model poisson/gaussian Default: "poisson"
    decoder_fr_clamp_min (float) - min value used when clamping decoded firing rates. Default: 1E-7
    decoder_fr_clamp_max (float) - max value used when clamping decoded firing rates. Default: 1E7
    z_prior_n_hidden_layers (int) - number of the ZPriorContinuous's MLP hidden layers. Default: 2
    z_prior_hidden_layer_dim (int) - dimension of the ZPriorContinuous's MLP hidden layers. Default: 20
    z_prior_hidden_layer_activation (nn.Module) - activation function applied to the ZPriorContinuous's MLP hidden layers. Default: nn.Tanh

    Notes
    -----
    decoder_affine_input_layer_slice_dim (int) - when None, x_dim // 2 is assigned. Otherwise assigns the specified value, assuming 0 < decoder_affine_input_layer_slice_dim < x_dim.
    decoder_affine_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(affine_hidden_layer_dim, x_dim // 4).
    decoder_nflow_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(nflow_hidden_layer_dim, x_dim // 4).
    """
    def __init__(
        self,
        x_dim: int,
        u_dim: int,
        z_dim: int,
        discrete_labels: bool = True,
        encoder_n_hidden_layers: int = 2,
        encoder_hidden_layer_dim: int = 120,
        encoder_hidden_layer_activation: nn.Module = nn.Tanh,
        decoder_n_gin_blocks: int = 2,
        decoder_gin_block_depth: int = 2,
        decoder_affine_input_layer_slice_dim: int = None,
        decoder_affine_n_hidden_layers: int = 2,
        decoder_affine_hidden_layer_dim: int = None,
        decoder_affine_hidden_layer_activation: nn.Module = nn.ReLU,
        decoder_nflow_n_hidden_layers: int = 2,
        decoder_nflow_hidden_layer_dim: int = None,
        decoder_nflow_hidden_layer_activation: nn.Module = nn.ReLU,
        decoder_observation_model: str = "poisson",
        decoder_fr_clamp_min: float = 1E-7,
        decoder_fr_clamp_max: float = 1E7,
        z_prior_n_hidden_layers: int = 2,
        z_prior_hidden_layer_dim: int = 20,
        z_prior_hidden_layer_activation: nn.Module = nn.Tanh
        ) -> None:
        super().__init__()

        if decoder_observation_model == "gaussian":
            self.observation_noise_model = nn.Linear(
                in_features=1,
                out_features=x_dim, 
                bias=False
            )
        elif decoder_observation_model == "poisson":
            self.observation_noise_model = None
        else:
            raise ValueError(f"Invalid observation model: {decoder_observation_model}")
        
        self.decoder = GINFlowDecoder(
            x_dim=x_dim,
            z_dim=z_dim,
            n_gin_blocks=decoder_n_gin_blocks,
            gin_block_depth=decoder_gin_block_depth,
            affine_input_layer_slice_dim=decoder_affine_input_layer_slice_dim,
            affine_n_hidden_layers=decoder_affine_n_hidden_layers,
            affine_hidden_layer_dim=decoder_affine_hidden_layer_dim,
            affine_hidden_layer_activation=decoder_affine_hidden_layer_activation,
            nflow_n_hidden_layers=decoder_nflow_n_hidden_layers,
            nflow_hidden_layer_dim=decoder_nflow_hidden_layer_dim,
            nflow_hidden_layer_activation=decoder_nflow_hidden_layer_activation,
            observation_model=decoder_observation_model
        )

        self.decoder_fr_clamp_min = decoder_fr_clamp_min
        self.decoder_fr_clamp_max = decoder_fr_clamp_max
        self.decoder_observation_model = decoder_observation_model
        
        self.encoder = MLPEncoder(
            x_dim=x_dim,
            z_dim=z_dim,
            n_hidden_layers=encoder_n_hidden_layers,
            hidden_layer_dim=encoder_hidden_layer_dim,
            activation=encoder_hidden_layer_activation
        )
        
        if discrete_labels:
            self.z_prior = ZPriorDiscrete(
                u_dim=u_dim,
                z_dim=z_dim
            )
        else:
            self.z_prior = ZPriorContinuous(
                u_dim=u_dim,
                z_dim=z_dim,
                n_hidden_layers=z_prior_n_hidden_layers,
                hidden_layer_dim=z_prior_hidden_layer_dim,
                hidden_layer_activation=z_prior_hidden_layer_activation
            )

    def forward(
        self,
        x: Tensor,
        u: Tensor
        ) -> Dict[str, Tensor]:
        
        # Mean and log of variance for each sample using label prior p(z|u)
        lambda_mean, lambda_log_variance = self.z_prior(u) 

        # Map each sample observation x to latent z approximating q(z|x)
        z_mean, z_log_variance = self.encoder(x)  

        # Compute the full posterior of q(z|x,u)~q(z|x)p(z|u) as a product of Gaussians
        posterior_mean, posterior_log_variance = compute_posterior(z_mean, z_log_variance, lambda_mean, lambda_log_variance)  

        # Sample latent z using reparameterization trick
        z_sample = generate_latent_z(
            mean=posterior_mean,
            log_variance=posterior_log_variance
        )  

        # Generate firing rate using sampled latent z
        firing_rate = self.decoder(z_sample)
        if self.decoder_observation_model == "poisson":
            firing_rate = clamp(firing_rate, min=self.decoder_fr_clamp_min, max=self.decoder_fr_clamp_max)

        return {
            "firing_rate": firing_rate,
            "lambda_mean": lambda_mean,
            "lambda_log_variance": lambda_log_variance,
            "posterior_mean": posterior_mean,
            "posterior_log_variance": posterior_log_variance,
            "z_mean": z_mean,
            "z_log_variance": z_log_variance,
            "z_sample": z_sample
        }



from typing import List

import torch
from torch import nn, Tensor


def build_mlp_layers(
    n_hidden_layers: int,
    hidden_layer_dims: List[int],
    activation: nn.Module
    ) -> nn.Sequential:
    """
    Helper method to construct a Multilayer Perceptron (MLP). 

    Parameters
    ----------
    n_hidden_layers (int) - the number of hidden in the model
    hidden_layer_dim (List[int]) - a list of integers specifying the input/output dimensions of each layer
    activation (nn.Module) - the activation function to apply after each hidden layer

    Returns
    -------
    nn.Sequential(*layers) - a Sequential container of the specified layers

    Notes
    -----
    A nn.Identity activation is applied after the final layer.
    """

    layers = []
    act = activation

    # Build layers
    for i in range(n_hidden_layers+1):
        # Output layer
        if i == n_hidden_layers:
            act = nn.Identity
        
        layers.append(nn.Linear(hidden_layer_dims[2*i], hidden_layer_dims[2*i+1]))
        layers.append(act())

    return nn.Sequential(*layers)

def compute_loss(
    x: Tensor,
    firing_rate: Tensor,
    lambda_mean: Tensor, 
    lambda_log_variance: Tensor,
    posterior_mean: Tensor, 
    posterior_log_variance: Tensor,
    observation_model: str,
    observation_noise_model: nn.Module = None
    ) -> Tensor:
    """ 
    pi-VAE Loss function

    Parameters
    ----------
    x (Tensor) - observed x. Size([n_samples, x_dim])
    firing_rate (Tensor) - decoded firing rates. Size([n_samples, x_dim])
    lambda_mean (Tensor) - means from label prior p(z|u). Size([n_samples, z_dim])
    lambda_log_variance (Tensor) - log of variances from label prior p(z|u). Size([n_samples, z_dim])
    posterior_mean (Tensor) - means from posterior q(z|x,u)~q(z|x)p(z|u). Size([n_samples. z_dim])
    posterior_log_variance (Tensor) - log of variances from posterior q(z|x,u)~q(z|x)p(z|u). Size([n_samples, z_dim])
    observation_model (str) - poisson or gaussian
    observation_noise_model (nn.Module) - if gaussian observation model, set the observation noise level as different real numbers. Default: None 
    
    Returns
    -------
    The total loss (Tensor)

    Notes
    -----
    observation_noise_model (nn.Module) - PiVAE uses an nn.Linear(1, x_dim, bias=False) Module

    min -log p(x|z) + E_q log(q(z))-log(p(z|u))
    cross entropy
    q (mean1, var1) p (mean2, var2)
    E_q log(q(z))-log(p(z|u)) = -0.5*(1-log(var2/var1) - (var1+(mean2-mean1)^2)/var2)
    E_q(z|x,u) log(q(z|x,u))-log(p(z|u)) = -0.5*(log(2*pi*var2) + (var1+(mean2-mean1)^2)/var2)
    p(z) = q(z|x) = N(f(x), g(x)) parametrized by nn
    """

    if observation_model == "poisson":
        observation_log_liklihood = torch.sum(firing_rate - x * torch.log(firing_rate), -1)
    elif observation_model == "gaussian":
        observation_log_variance = observation_noise_model(torch.ones((1, 1)))
        observation_log_liklihood = torch.sum(torch.square(firing_rate - x) / (2 * torch.exp(observation_log_variance)) + (observation_log_variance / 2), -1)
    else:
        raise ValueError(f"Invalid observation model: {observation_model}")

    loss = 1 + posterior_log_variance - lambda_log_variance - ((torch.square(posterior_mean - lambda_mean) + torch.exp(posterior_log_variance)) / torch.exp(lambda_log_variance))
    kl_loss = 0.5 * torch.sum(loss, dim=-1)

    return torch.mean(observation_log_liklihood - kl_loss)

def compute_posterior(
    z_mean: Tensor, 
    z_log_variance: Tensor, 
    lambda_mean: Tensor, 
    lambda_log_variance: Tensor,
    ) -> Tensor:
    """
    Compute the full posterior of q(z|x,u)~q(z|x)p(z|u) as a product of Gaussians.

    Parameters
    ----------
    z_mean (Tensor) - means of encoded distribution q(z|x). Size([n_samples, z_dim])
    z_log_variance (Tensor) - log of varinces of encoded distribution q(z|x). Size([n_samples, z_dim])
    lambda_mean (Tensor) - means of label prior distribution p(z|u). Size([n_samples, z_dim])
    lambda_log_variance (Tensor) - log of variances of label prior distribution p(z|u). Size([n_samples, z_dim])

    Returns
    -------
    posterior_mean: approximate posterior means of distribution q(z|x,u). Size([n_samples, z_dim])
    posterior_log_variance: approximate posterior log of variances of distribution q(z|x,u). Size([n_samples, z_dim])
    """

    variance_difference = z_log_variance - lambda_log_variance
    posterior_mean = (z_mean / (1 + torch.exp(variance_difference))) + (lambda_mean / (1 + torch.exp(torch.neg(variance_difference))))
    posterior_log_variance = z_log_variance + lambda_log_variance - torch.log(torch.exp(z_log_variance) + torch.exp(lambda_log_variance))

    return posterior_mean, posterior_log_variance

def generate_latent_z(
    mean: Tensor,
    log_variance: Tensor
    ) -> Tensor:
    """
    Reparameterization trick by sampling from an isotropic unit Gaussian.

    Parameters
    -----------
    mean (Tensor) - means of each z sample. Size([n_samples, z_dim])
    log_variance (Tensor) - log of variances of each z sample. Size([n_samples, z_dim])

    Returns
    ------
    Samples of latent z. Size([n_samples, z_dim])
    """

    return mean + torch.exp(0.5 * log_variance) * torch.randn_like(mean)
